@@ -12,10 +12,12 @@ import json
 import logging
 import os
 import re
+import shlex
 import subprocess
 import time
 import urllib.error
 import urllib.request
+import uuid
 from dataclasses import dataclass, field
 from pathlib import Path
 from typing import Dict, List, Literal, Optional
@@ -24,12 +26,41 @@ logger = logging.getLogger("selftest")
 
 # pylint: disable=broad-exception-caught
 # pylint: disable=line-too-long
-# pylint: disable=too-many-lines
+# pylint: disable=too-many-branches
 # pylint: disable=too-many-instance-attributes
 # pylint: disable=too-many-lines
 # pylint: disable=too-many-locals
 
+SYS_CLASS_BLOCK = "/sys/class/block"
 SYS_CLASS_NET = "/sys/class/net"
+AZURE_EPHEMERAL_DISK_SETUP_CONF = "/etc/azure-ephemeral-disk-setup.conf"
+AZURE_EPHEMERAL_DISK_SETUP_SERVICE = "azure-ephemeral-disk-setup.service"
+DEV_DISK_AZURE_RESOURCE = "/dev/disk/azure/resource"
+DEV_DISK_CLOUD_AZURE_RESOURCE = "/dev/disk/cloud/azure_resource"
+
+
+def is_systemd_fixed_for_unmanaged_bug(systemd_version: str) -> bool:
+    """Check if systemd version is >= 257.7.
+
+    Due to https://github.com/systemd/systemd/issues/36997 there may be
+    an unexpected IP assigned to VF that we would otherwise assert on.
+    """
+    parts = systemd_version.split(".")
+    match = re.search(r"\d+", parts[0])
+    if not match:
+        raise RuntimeError(
+            f"Unexpected systemd version format: {systemd_version}. Expected format is 'major.minor'."
+        )
+    major = int(match.group())
+    minor = 0
+
+    # minor may have -patch data, remove it then convert to int.
+    if len(parts) > 1:
+        match = re.search(r"\d+", parts[1])
+        if match:
+            minor = int(match.group())
+
+    return major > 257 or (major == 257 and minor >= 7)
 
 
 @dataclass(eq=True, repr=True)
@@ -63,6 +94,12 @@ def gb_to_gib(size_gb: int) -> int:
     return int(size_gb * (1000**3) / (1024**3))
 
 
+def unchecked_run(cmd) -> str:
+    """ "Run a command without checking the return code and return stripped output."""
+    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
+    return result.stdout.strip()
+
+
 SKU_CONFIGS = {
     "Standard_B2ts_v2": SkuConfig(vm_size="Standard_B2ts_v2"),
     # "Standard_D2s_v3": SkuConfig(vm_size="Standard_D2s_v3", temp_disk_size_gib=16),
@@ -187,6 +224,24 @@ def device_sort(devices: List[str]) -> List[str]:
     return sorted(devices, key=natural_sort_key)
 
 
+def get_disk_io_timeout(disk_name: str) -> int:
+    """Get disk's queue/io_timeout."""
+    timeout_path = Path(SYS_CLASS_BLOCK, disk_name, "queue/io_timeout")
+    logger.debug("checking disk %s io_timeout: %s", disk_name, timeout_path)
+
+    if not timeout_path.exists():
+        logger.error("disk %s io_timeout not found", disk_name)
+        raise FileNotFoundError(f"disk {disk_name} io_timeout not found")
+
+    timeout = timeout_path.read_text(encoding="utf-8").strip()
+
+    try:
+        return int(timeout)
+    except ValueError:
+        logger.error("invalid value for disk %s io_timeout: %r", disk_name, timeout)
+        raise
+
+
 def get_disk_size_gb(disk_path: str) -> int:
     """Get the size of the disk in GB."""
     try:
@@ -234,7 +289,8 @@ def get_disk_size_gib(disk_path: str) -> int:
 def get_imds_metadata() -> Dict:
     """Fetch IMDS metadata using urllib."""
     url = "http://169.254.169.254/metadata/instance?api-version=2021-02-01"
-    headers = {"Metadata": "true"}
+    request_id = str(uuid.uuid4())
+    headers = {"Metadata": "true", "x-ms-client-request-id": request_id}
 
     req = urllib.request.Request(url, headers=headers)
 
@@ -252,11 +308,15 @@ def get_imds_metadata() -> Dict:
                         None,
                     )
                 metadata = json.load(response)
-                logger.debug("fetched IMDS metadata: %r", metadata)
+                logger.debug(
+                    "fetched IMDS metadata with request id %s: %r", request_id, metadata
+                )
                 return metadata
         except urllib.error.URLError as error:
             last_error = error
-            logger.error("error fetching IMDS metadata: %r", error)
+            logger.error(
+                "error fetching IMDS metadata with request id %s: %r", request_id, error
+            )
             time.sleep(1)
 
     raise RuntimeError(f"failed to fetch IMDS metadata: {last_error}")
@@ -372,6 +432,8 @@ def get_root_block_device() -> str:
 def get_scsi_resource_disk() -> Optional[str]:
     """Get the SCSI resource disk device."""
     paths = [
+        # azure resource disk
+        "/dev/disk/azure/resource",
         # cloud-init udev rules
         "/dev/disk/cloud/azure_resource",
         # gen2
@@ -397,9 +459,6 @@ def get_scsi_resource_disk() -> Optional[str]:
     return None
 
 
-DEV_DISK_AZURE_RESOURCE = "/dev/disk/azure/resource"
-
-
 @dataclass(eq=True, repr=True)
 class DiskInfo:
     """Information about different types of disks present."""
@@ -408,6 +467,8 @@ class DiskInfo:
     dev_disk_azure_links: List[str] = field(default_factory=list)
     dev_disk_azure_resource_disk: Optional[str] = None  # resolved path
     dev_disk_azure_resource_disk_size_gib: int = 0
+    dev_disk_cloud_azure_resource: Optional[str] = None  # resolved path
+    nvme_io_timeouts: Dict[str, int] = field(default_factory=dict)
     nvme_local_disk_size_gib: int = 0
     nvme_local_disks_v1: List[str] = field(default_factory=list)
     nvme_local_disks_v2: List[str] = field(default_factory=list)
@@ -440,6 +501,12 @@ class DiskInfo:
                 dev_disk_azure_resource_disk
             )
 
+        dev_disk_cloud_azure_resource = None
+        if os.path.exists(DEV_DISK_CLOUD_AZURE_RESOURCE):
+            dev_disk_cloud_azure_resource = os.path.realpath(
+                DEV_DISK_CLOUD_AZURE_RESOURCE
+            )
+
         nvme_local_disks_v1 = get_nvme_namespace_devices_with_model(
             "Microsoft NVMe Direct Disk"
         )
@@ -461,8 +528,8 @@ class DiskInfo:
 
         nvme_remote_disks = get_remote_nvme_disks()
         if nvme_remote_disks:
-            nvme_remote_os_disk = nvme_remote_disks.pop(0)
-            nvme_remote_data_disks = nvme_remote_disks
+            nvme_remote_os_disk = nvme_remote_disks[0]
+            nvme_remote_data_disks = nvme_remote_disks[1:]
         else:
             nvme_remote_os_disk = None
             nvme_remote_data_disks = []
@@ -476,10 +543,16 @@ class DiskInfo:
             get_disk_size_gib(f"/dev/{scsi_resource_disk}") if scsi_resource_disk else 0
         )
 
+        nvme_io_timeouts: Dict[str, int] = {}
+        for disk_name in nvme_local_disks + nvme_remote_disks:
+            nvme_io_timeouts[disk_name] = get_disk_io_timeout(disk_name)
+
         disk_info = cls(
             dev_disk_azure_links=dev_disk_azure_links,
             dev_disk_azure_resource_disk=dev_disk_azure_resource_disk,
             dev_disk_azure_resource_disk_size_gib=dev_disk_azure_resource_disk_size_gib,
+            dev_disk_cloud_azure_resource=dev_disk_cloud_azure_resource,
+            nvme_io_timeouts=nvme_io_timeouts,
             nvme_local_disk_size_gib=nvme_local_disk_size_gib,
             nvme_local_disks_v1=nvme_local_disks_v1,
             nvme_local_disks_v2=nvme_local_disks_v2,
@@ -493,10 +566,338 @@ class DiskInfo:
             scsi_resource_disk_size_gib=scsi_resource_disk_size_gib,
         )
 
-        logger.info("disks info: %r", disk_info)
         return disk_info
 
 
+@dataclass(eq=True, repr=True)
+class AzureEphemeralDiskConfig:
+    """Parsed /etc/azure-ephemeral-disk-setup.conf config."""
+
+    aggregation: str = "mdadm"
+    fs_type: str = "ext4"
+    mdadm_chunk: str = "512K"
+    mdadm_name: str = "azure-ephemeral-md"
+    mount_point: str = "/mnt"
+    scsi_resource: bool = False
+
+    @staticmethod
+    def _parse_bool(value: str) -> Optional[bool]:
+        if value == "true":
+            return True
+        if value == "false":
+            return False
+        return None
+
+    @classmethod
+    def gather(cls) -> "AzureEphemeralDiskConfig":
+        """Parse service config."""
+        aggregation = "auto"
+        fs_type = "ext4"
+        mdadm_chunk = "512K"
+        mdadm_name = "azure-ephemeral-md"
+        mount_point = "/mnt"
+        scsi_resource = False
+
+        config_path = Path(AZURE_EPHEMERAL_DISK_SETUP_CONF)
+        if not config_path.exists():
+            return cls()
+
+        config_str = config_path.read_text(encoding="utf-8")
+        for line in config_str.splitlines():
+            line = line.strip()
+            if not line or line.startswith("#"):
+                continue
+
+            key, value = line.split("=", 1)
+            value = shlex.split(value, posix=True)[0]
+            if key == "AZURE_EPHEMERAL_DISK_SETUP_AGGREGATION":
+                if value not in {"auto", "mdadm", "none"}:
+                    raise ValueError(f"Invalid aggregation: {value}")
+                aggregation = value
+            elif key == "AZURE_EPHEMERAL_DISK_SETUP_FS_TYPE":
+                if value not in {"ext4", "xfs"}:
+                    raise ValueError(f"Invalid filesystem type: {value}")
+                fs_type = value
+            elif key == "AZURE_EPHEMERAL_DISK_SETUP_MDADM_CHUNK":
+                if not re.match(r"^\d+[KMGT]$", value, re.IGNORECASE):
+                    raise ValueError(f"Invalid mdadm chunk size: {value}")
+                mdadm_chunk = value
+            elif key == "AZURE_EPHEMERAL_DISK_SETUP_MDADM_NAME":
+                if not re.match(r"^[a-zA-Z0-9_-]+$", value):
+                    raise ValueError(f"Invalid mdadm array name: {value}")
+                mdadm_name = value
+            elif key == "AZURE_EPHEMERAL_DISK_SETUP_MOUNT_POINT":
+                if not os.path.isabs(value) or " " in value:
+                    raise ValueError(f"Invalid mount point: {value}")
+                mount_point = value
+            elif key == "AZURE_EPHEMERAL_DISK_SETUP_SCSI_RESOURCE":
+                value_bool = cls._parse_bool(value)
+                if value_bool is None:
+                    raise ValueError(f"Invalid boolean for scsi_resource: {value}")
+                scsi_resource = value_bool
+            else:
+                raise ValueError(
+                    f"Unexpected key={key} line in {AZURE_EPHEMERAL_DISK_SETUP_CONF}: {line}"
+                )
+
+        return cls(
+            aggregation=aggregation,
+            fs_type=fs_type,
+            mdadm_chunk=mdadm_chunk,
+            mdadm_name=mdadm_name,
+            mount_point=mount_point,
+            scsi_resource=scsi_resource,
+        )
+
+
+@dataclass(eq=True, repr=True)
+class ServiceInfo:
+    """Status of various system services."""
+
+    systemd_version: str
+    systemd_version_fixed_unmanaged_bug: bool
+    cloud_init_service_enabled: bool
+    ephemeral_service_enabled: bool
+    ephemeral_service_active: bool
+    ephemeral_service_failed: bool
+    ephemeral_service_journal: Optional[str] = None
+    waagent_resource_disk_format: Optional[bool] = None
+
+    @classmethod
+    def gather(cls) -> "ServiceInfo":
+        """Gather information about necessary system services."""
+        cloud_init_service_enabled = (
+            unchecked_run(["systemctl", "is-enabled", "cloud-init-local.service"])
+            == "enabled"
+        )
+        ephemeral_service_enabled = (
+            unchecked_run(
+                ["systemctl", "is-enabled", AZURE_EPHEMERAL_DISK_SETUP_SERVICE]
+            )
+            == "enabled"
+        )
+        ephemeral_service_active = (
+            unchecked_run(
+                ["systemctl", "is-active", AZURE_EPHEMERAL_DISK_SETUP_SERVICE]
+            )
+            == "active"
+        )
+        ephemeral_service_failed = (
+            unchecked_run(
+                ["systemctl", "is-failed", AZURE_EPHEMERAL_DISK_SETUP_SERVICE]
+            )
+            == "failed"
+        )
+        ephemeral_service_journal = unchecked_run(
+            [
+                "journalctl",
+                "--no-pager",
+                "--output=short-precise",
+                "--boot",
+                "--unit",
+                AZURE_EPHEMERAL_DISK_SETUP_SERVICE,
+            ]
+        )
+
+        # Get systemd version line. Then parse for detailed version.
+        # Example first line output: `systemd 255 (255.4-1ubuntu8.8)`.
+        systemd_version = (
+            unchecked_run(["systemctl", "--version"]).strip().split("\n")[0]
+        )
+        systemd_version = systemd_version.split()[-1].strip("()")
+        systemd_version_fixed_unmanaged_bug = is_systemd_fixed_for_unmanaged_bug(
+            systemd_version
+        )
+
+        waagent_config_path = Path("/etc/waagent.conf")
+        waagent_resource_disk_format = (
+            "ResourceDisk.Format=y" in waagent_config_path.read_text(encoding="utf-8")
+            if waagent_config_path.exists()
+            else None
+        )
+
+        return cls(
+            cloud_init_service_enabled=cloud_init_service_enabled,
+            ephemeral_service_enabled=ephemeral_service_enabled,
+            ephemeral_service_active=ephemeral_service_active,
+            ephemeral_service_failed=ephemeral_service_failed,
+            ephemeral_service_journal=ephemeral_service_journal,
+            systemd_version=systemd_version,
+            systemd_version_fixed_unmanaged_bug=systemd_version_fixed_unmanaged_bug,
+            waagent_resource_disk_format=waagent_resource_disk_format,
+        )
+
+
+@dataclass(eq=True, repr=True)
+class Mount:
+    """Mount parameters."""
+
+    target: str
+    source: Optional[str] = None
+    fstype: Optional[str] = None
+    options: Optional[str] = None
+    rest: Dict[str, str] = field(default_factory=dict)
+
+
+@dataclass(eq=True, repr=True)
+class MountInfo:
+    """Information about mounted filesystems from kernel and fstab sources."""
+
+    fstab_mounts: Dict[str, Mount] = field(default_factory=dict)
+    kernel_mounts: Dict[str, Mount] = field(default_factory=dict)
+
+    @classmethod
+    def gather(cls) -> "MountInfo":
+        """Gather mount info from both kernel and fstab."""
+        return cls(
+            kernel_mounts=cls._gather_from_findmnt(mode="kernel"),
+            fstab_mounts=cls._gather_from_findmnt(mode="fstab"),
+        )
+
+    @staticmethod
+    def _gather_from_findmnt(mode: str) -> Dict[str, Mount]:
+        """Helper to run findmnt --json --<mode> and parse output."""
+        assert mode in ("fstab", "kernel")
+        cmd = ["findmnt", "--json", f"--{mode}"]
+        proc = subprocess.run(cmd, capture_output=True, check=False, text=True)
+        if proc.returncode != 0:
+            logger.error(
+                "error running findmnt: %r (stderr=%r stdout=%r)",
+                proc,
+                proc.stderr,
+                proc.stdout,
+            )
+            return {}
+
+        data = json.loads(proc.stdout)
+
+        mounts: Dict[str, Mount] = {}
+
+        def add_mounts(filesystems):
+            for fs in filesystems:
+                target = fs.get("target")
+                if not target:
+                    continue
+
+                mounts[target] = Mount(
+                    target=target,
+                    source=fs.get("source"),
+                    fstype=fs.get("fstype"),
+                    options=fs.get("options"),
+                    rest={
+                        k: v
+                        for k, v in fs.items()
+                        if k
+                        not in {"target", "source", "fstype", "options", "children"}
+                    },
+                )
+
+                # Recurse into children if they exist
+                children = fs.get("children", [])
+                if isinstance(children, list):
+                    add_mounts(children)
+
+        add_mounts(data.get("filesystems", []))
+        return mounts
+
+    def validate_ephemeral_mount(
+        self,
+        disk_info: DiskInfo,
+        azure_ephemeral_disk_config: AzureEphemeralDiskConfig,
+        service_info: ServiceInfo,
+    ) -> None:
+        """Validate ephemeral disk mount."""
+        mnt_fstab = self.fstab_mounts.get("/mnt")
+        mnt_kernel = self.kernel_mounts.get("/mnt")
+
+        cloud_init_resource_disk_link_exists = (
+            disk_info.dev_disk_cloud_azure_resource is not None
+        )
+        if service_info.ephemeral_service_enabled:
+            # The service is expected to fail if is configured to manage the
+            # resource disk at the same time cloud-init is, or WALA is configured
+            # to manage the resource disk.
+            if azure_ephemeral_disk_config.scsi_resource and (
+                cloud_init_resource_disk_link_exists
+                or service_info.waagent_resource_disk_format
+            ):
+                assert (
+                    service_info.ephemeral_service_failed
+                ), f"azure-ephemeral-disk-setup service did not fail as expected: {service_info.ephemeral_service_journal}"
+            else:
+                assert (
+                    service_info.ephemeral_service_active
+                ), f"azure-ephemeral-disk-setup service is enabled but not active: {service_info.ephemeral_service_journal}"
+                assert (
+                    not service_info.ephemeral_service_failed
+                ), f"azure-ephemeral-disk-setup service is enabled but failed: {service_info.ephemeral_service_journal}"
+
+        mount_expected = (
+            (
+                disk_info.scsi_resource_disk
+                and cloud_init_resource_disk_link_exists
+                and service_info.cloud_init_service_enabled
+            )
+            or (
+                disk_info.scsi_resource_disk
+                and azure_ephemeral_disk_config.scsi_resource
+                and service_info.ephemeral_service_active
+            )
+            or (disk_info.nvme_local_disks and service_info.ephemeral_service_active)
+        )
+        if mount_expected:
+            assert (
+                mnt_fstab and mnt_fstab.options and mnt_fstab.source
+            ), f"no fstab mount found for /mnt: {self.fstab_mounts}"
+            assert (
+                mnt_kernel and mnt_kernel.options and mnt_kernel.source
+            ), f"no kernel mount found for /mnt: {self.kernel_mounts}"
+
+            # Assume cloud-init manages SCSI resource disk if cloud-init's symlink is present.
+            if (
+                cloud_init_resource_disk_link_exists
+                and service_info.cloud_init_service_enabled
+            ):
+                assert (
+                    "comment=cloudconfig" in mnt_fstab.options
+                ), f"expected comment=cloudconfig in fstab options for /mnt: {mnt_fstab.options}"
+            else:
+                assert (
+                    mnt_fstab.options
+                    == "defaults,nofail,comment=azure-ephemeral-disk-setup"
+                ), f"unexpected fstab options for /mnt: {mnt_fstab.options}"
+
+                assert (
+                    mnt_fstab.source == "LABEL=AzureEphmDsk"
+                ), f"unexpected fstab source for /mnt: {mnt_fstab.source}"
+
+            assert mnt_kernel.options and (
+                "rw" in mnt_kernel.options
+            ), f"missing rw in fstab options: {mnt_kernel.options}"
+        else:
+            if mnt_fstab:
+                assert mnt_fstab.options and (
+                    "comment=cloudconfig" in mnt_fstab.options
+                ), f"unexpected fstab mount for /mnt: {mnt_fstab}"
+
+            assert not mnt_kernel, f"unexpected kernel mount for /mnt: {mnt_kernel}"
+
+        logger.info(
+            "validate_ephemeral_mount OK: fstab=%r kernel=%r", mnt_fstab, mnt_kernel
+        )
+
+    def validate(
+        self,
+        disk_info: DiskInfo,
+        azure_ephemeral_disk_config: AzureEphemeralDiskConfig,
+        service_info: ServiceInfo,
+    ) -> None:
+        """Validate mounts."""
+        self.validate_ephemeral_mount(
+            disk_info, azure_ephemeral_disk_config, service_info
+        )
+
+
 @dataclass
 class AzureNvmeIdDevice:
     """Azure NVMe ID device."""
@@ -594,7 +995,7 @@ class AzureNvmeIdInfo:
             assert disk_cfg.nvme_id, f"unexpected remote disk id {disk_cfg}"
             assert not disk_cfg.extra, f"unexpected remote disk extra {disk_cfg}"
 
-        logger.info("validate_azure_nvme_disks OK: %r", self.azure_nvme_id_disks)
+        logger.info("validate_azure_nvme_disks OK")
 
     def validate_azure_nvme_id(self, disk_info: DiskInfo) -> None:
         """Validate azure-nvme-id outputs."""
@@ -610,7 +1011,7 @@ class AzureNvmeIdInfo:
             ), f"unexpected azure-nvme-id stderr: {self.azure_nvme_id_stderr}"
 
         self._validate_azure_nvme_disks(self.azure_nvme_id_disks, disk_info)
-        logger.info("validate_azure_nvmve_id OK: %r", self.azure_nvme_id_stdout)
+        logger.info("validate_azure_nvme_id OK")
 
     def validate_azure_nvme_id_help(self) -> None:
         """Validate azure-nvme-id --help outputs."""
@@ -623,9 +1024,7 @@ class AzureNvmeIdInfo:
             and self.azure_nvme_id_help_stdout.startswith("Usage: azure-nvme-id ")
         ), "unexpected azure-nvme-id --help stdout: {self.azure_nvme_id_help_stdout!r}"
 
-        logger.info(
-            "validate_azure_nvme_id_help OK: %r", self.azure_nvme_id_help_stdout
-        )
+        logger.info("validate_azure_nvme_id_help OK")
 
     def validate_azure_nvme_id_json(self, disk_info: DiskInfo) -> None:
         """Validate azure-nvme-id --format json outputs."""
@@ -651,9 +1050,7 @@ class AzureNvmeIdInfo:
             )
             for disk in self.azure_nvme_id_json_disks.values()
         ), "missing model in azure-nvme-id --format json"
-        logger.info(
-            "validate_azure_nvmve_id_json OK: %r", self.azure_nvme_id_json_stdout
-        )
+        logger.info("validate_azure_nvme_id_json OK")
 
     def validate_azure_nvme_id_version(self) -> None:
         """Validate azure-nvme-id --version outputs."""
@@ -686,9 +1083,7 @@ class AzureNvmeIdInfo:
             and self.azure_nvme_id_zzz_stdout.startswith("Usage: azure-nvme-id ")
         ), (f"unexpected azure-nvme-id zzz stdout: {self.azure_nvme_id_zzz_stdout!r}")
 
-        logger.info(
-            "validate_azure_nvme_id_invalid_arg OK: %r", self.azure_nvme_id_zzz_stdout
-        )
+        logger.info("validate_azure_nvme_id_invalid_arg OK")
 
     def validate(self, disk_info: DiskInfo) -> None:
         """Validate Azure NVMe ID output."""
@@ -761,7 +1156,7 @@ class AzureNvmeIdInfo:
             azure_nvme_id_zzz_stdout=azure_nvme_id_zzz_stdout,
             azure_nvme_id_zzz_stderr=azure_nvme_id_zzz_stderr,
         )
-        logger.info("azure-nvme-id info: %r", azure_nvme_id_info)
+
         return azure_nvme_id_info
 
     @staticmethod
@@ -976,7 +1371,9 @@ class NetworkInfo:
             )
             return {}
 
-    def _validate_interface(self, interface: NetworkInterface) -> None:
+    def _validate_interface(
+        self, interface: NetworkInterface, service_info: ServiceInfo
+    ) -> None:
         """Ensure the required properties are set for hv_netvsc, mlx4, mlx5, and mana devices."""
         if interface.driver in ["mlx4_core", "mlx5_core", "mana"]:
             assert (
@@ -1001,16 +1398,31 @@ class NetworkInfo:
         ):
             assert interface.ipv4_addrs, f"missing IPv4 addresses for {interface}"
         else:
-            assert (
-                not interface.ipv4_addrs
-            ), f"unexpected IPv4 addresses for {interface}"
+            # Ensure accelerated network VFs do not have an IP assigned.  There is a
+            # systemd bug which will cause this check to fail, only log the failure
+            # if unpatched to avoid unnecessary assertion failures.
+            if service_info.systemd_version_fixed_unmanaged_bug:
+                assert (
+                    not interface.ipv4_addrs
+                ), f"unexpected IPv4 addresses for {interface}"
+            elif interface.ipv4_addrs:
+                logger.error(
+                    "unexpected IPv4 addresses for %s: %r",
+                    interface.name,
+                    interface.ipv4_addrs,
+                )
 
-        logger.info("validate_interface %s OK: %r", interface.name, interface)
+        logger.info(
+            "validate_interface %s %s OK: %r",
+            interface.name,
+            interface.driver,
+            interface.ipv4_addrs,
+        )
 
-    def validate(self) -> None:
+    def validate(self, service_info: ServiceInfo) -> None:
         """Validate network configuration."""
         for _, interface in self.interfaces.items():
-            self._validate_interface(interface)
+            self._validate_interface(interface, service_info)
 
     @classmethod
     def gather(cls) -> "NetworkInfo":
@@ -1025,15 +1437,16 @@ class AzureVmUtilsValidator:
         self,
         *,
         skip_imds_validation: bool = False,
-        skip_network_validation: bool = False,
-        skip_symlink_validation: bool = False,
+        skip_udev_validation: bool = False,
     ) -> None:
+        self.azure_ephemeral_disk_config = AzureEphemeralDiskConfig.gather()
         self.azure_nvme_id_info = AzureNvmeIdInfo.gather()
         self.disk_info = DiskInfo.gather()
+        self.mount_info = MountInfo.gather()
         self.net_info = NetworkInfo.gather()
+        self.service_info = ServiceInfo.gather()
         self.skip_imds_validation = skip_imds_validation
-        self.skip_network_validation = skip_network_validation
-        self.skip_symlink_validation = skip_symlink_validation
+        self.skip_udev_validation = skip_udev_validation
 
         try:
             self.imds_metadata = get_imds_metadata()
@@ -1047,6 +1460,12 @@ class AzureVmUtilsValidator:
         self.sku_config = SKU_CONFIGS.get(self.vm_size)
 
         logger.info("sku config: %r", self.sku_config)
+        logger.info("azure_ephemeral_disk_config: %r", self.azure_ephemeral_disk_config)
+        logger.info("azure_nvme_id_info: %r", self.azure_nvme_id_info)
+        logger.info("disk_info: %r", self.disk_info)
+        logger.info("mount_info: %r", self.mount_info)
+        logger.info("net_info: %r", self.net_info)
+        logger.info("service_info: %r", self.service_info)
 
     def validate_dev_disk_azure_links_data(self) -> None:
         """Validate /dev/disk/azure/data links.
@@ -1070,9 +1489,18 @@ class AzureVmUtilsValidator:
                 self.disk_info.nvme_remote_data_disks
             ), f"unexpected number of data disks: {data_disks} configured={self.disk_info.nvme_remote_data_disks}"
 
-        assert (
-            len(data_disks) == expected_data_disks
-        ), f"unexpected number of data disks: {data_disks} IMDS configured={imds_data_disks} (note that IMDS may not be accurate)"
+        if expected_data_disks == 0 and len(data_disks) > 0:
+            # Treat this as a soft error because IMDS metadata may be missing details
+            # that we would need to retry on for some unknown amount of time.
+            logger.error(
+                "IMDS reports no data disks but /dev/disk/azure/data/by-lun links found: %r"
+                " (assuming IMDS metadata is invalid)",
+                data_disks,
+            )
+        else:
+            assert (
+                len(data_disks) == expected_data_disks
+            ), f"unexpected number of data disks: {data_disks} IMDS configured={imds_data_disks} (note that IMDS may not be accurate)"
 
         # Verify disk sizes match up with IMDS configuration.
         for imds_disk in imds_data_disks:
@@ -1155,7 +1583,7 @@ class AzureVmUtilsValidator:
 
     def validate_dev_disk_azure_links_resource(self) -> None:
         """Validate /dev/disk/azure/resource link."""
-        resource_disk = "/dev/disk/azure/resource"
+        resource_disk = DEV_DISK_AZURE_RESOURCE
         expected = (self.sku_config and self.sku_config.temp_disk_size_gib) or bool(
             self.disk_info.scsi_resource_disk
         )
@@ -1170,10 +1598,30 @@ class AzureVmUtilsValidator:
 
         logger.info("validate_dev_disk_azure_links_resource OK: %r", resource_disk)
 
+    def validate_mounts(self) -> None:
+        """Ensure SCSI resource disk and/or NVMe local disks are mounted correctly."""
+        self.mount_info.validate(
+            self.disk_info, self.azure_ephemeral_disk_config, self.service_info
+        )
+
     def validate_networking(self) -> None:
         """Validate networking configuration."""
-        self.net_info.validate()
-        logger.info("validate_networking OK: %r", self.net_info)
+        self.net_info.validate(self.service_info)
+        logger.info("validate_networking OK")
+
+    def validate_nvme_io_timeouts(self) -> None:
+        """Validate NVMe queue I/O timeouts."""
+        for namespace in self.disk_info.nvme_remote_disks:
+            assert (
+                namespace in self.disk_info.nvme_io_timeouts
+            ), f"missing NVMe I/O timeout for {namespace}"
+
+            timeout = self.disk_info.nvme_io_timeouts[namespace]
+            assert (
+                timeout == 240000
+            ), f"unexpected NVMe I/O timeout for {namespace}: {timeout} (expected 240000 ms)"
+
+        logger.info("validate_nvme_io_timeouts OK: %r", self.disk_info.nvme_io_timeouts)
 
     def validate_nvme_local_disks(self) -> None:
         """Validate NVMe local disks."""
@@ -1201,6 +1649,10 @@ class AzureVmUtilsValidator:
             self.disk_info.dev_disk_azure_resource_disk,
         )
 
+    def validate_services(self) -> None:
+        """Validate services."""
+        logger.info("validate_services OK")
+
     def validate_sku_config(self) -> None:
         """Validate SKU config."""
         if not self.sku_config:
@@ -1230,29 +1682,33 @@ class AzureVmUtilsValidator:
             == self.sku_config.temp_disk_size_gib
         ), f"temp disk size mismatch: {self.disk_info.dev_disk_azure_resource_disk_size_gib} != {self.sku_config.temp_disk_size_gib}"
 
-        logger.info("validate_sku_config OK: %r", self.sku_config)
+        logger.info("validate_sku_config OK")
 
     def validate(self) -> None:
         """Run validations."""
+        logger.info("Starting validation...")
         self.azure_nvme_id_info.validate(self.disk_info)
 
-        if self.skip_symlink_validation:
+        if self.skip_udev_validation:
             logger.info("validate_dev_disk_azure_links_data SKIPPED")
             logger.info("validate_dev_disk_azure_links_local SKIPPED")
             logger.info("validate_dev_disk_azure_links_os SKIPPED")
             logger.info("validate_dev_disk_azure_links_resource SKIPPED")
+            logger.info("validate_mounts SKIPPED")
             logger.info("validate_scsi_resource_disk SKIPPED")
+            logger.info("validate_services SKIPPED")
+            logger.info("validate_networking SKIPPED")
+            logger.info("validate_nvme_io_timeouts SKIPPED")
         else:
             self.validate_dev_disk_azure_links_data()
             self.validate_dev_disk_azure_links_local()
             self.validate_dev_disk_azure_links_os()
             self.validate_dev_disk_azure_links_resource()
-            self.validate_scsi_resource_disk()
-
-        if self.skip_network_validation:
-            logger.info("validate_networking SKIPPED")
-        else:
+            self.validate_mounts()
             self.validate_networking()
+            self.validate_nvme_io_timeouts()
+            self.validate_services()
+            self.validate_scsi_resource_disk()
 
         self.validate_sku_config()
 
@@ -1275,14 +1731,9 @@ def main() -> None:
         help="Skip imds validation (allow for running tests outside Azure VM)",
     )
     parser.add_argument(
-        "--skip-network-validation",
+        "--skip-udev-validation",
         action="store_true",
-        help="Skip network validation (allow for running test without reboot after install)",
-    )
-    parser.add_argument(
-        "--skip-symlink-validation",
-        action="store_true",
-        help="Skip symlink validation (allow for running test without reboot after install)",
+        help="Skip udev validation (allow for running test without reboot after install)",
     )
     args = parser.parse_args()
 
@@ -1293,8 +1744,7 @@ def main() -> None:
 
     validator = AzureVmUtilsValidator(
         skip_imds_validation=args.skip_imds_validation,
-        skip_network_validation=args.skip_network_validation,
-        skip_symlink_validation=args.skip_symlink_validation,
+        skip_udev_validation=args.skip_udev_validation,
     )
     validator.validate()
 
